//
//  PKWebSocketFrameParser-Hixie76.m
//  PKWebSocket
//
//  Created by Pavel Kunc on 20/04/2012.
//  Copyright (c) 2012 Pavel Kunc. All rights reserved.
//

#import "PKWebSocketFrameParser-Hixie76.h"

@implementation PKWebSocketFrameParserHixie76 {
    NSData *_textFrameEnd;
    NSData *_textFrameStart;
}

- (id)init {
    self = [super init];
    if (self != nil) {
        self->_textFrameStart = [NSData dataWithBytes:"\x00" length:1];
        self->_textFrameEnd = [NSData dataWithBytes:"\xFF" length:1];
        self.closingPayload = [NSData dataWithBytes:"\xFF\x00" length:2];
    }
    return self;
}

- (NSData *)textFramePayloadWithString:(NSString *)string {
    NSMutableData *payload = [NSMutableData dataWithData:_textFrameStart];
    [payload appendData:[string dataUsingEncoding:NSUTF8StringEncoding]];
    [payload appendData:_textFrameEnd];
    return (NSData *)payload;
}

- (void)processData:(NSData *)aData {
    // Make sure we have delegate
    NSAssert(self.delegate, @"No delegate set!");

    @autoreleasepool {
        if (aData == nil || [aData length] == 0) return;

        char frameType;
        BOOL finished = NO;
        NSUInteger length = 0;
        NSUInteger pointer = 0;

        while (!finished) {
            // Set frame type
            [aData getBytes:&frameType range:NSMakeRange(pointer,1)];
            pointer++;

            if ((frameType & 0x80) == 0x80) {
                // High order bit set
                char b;
                do {
                    // If we don't have more data get out before exception
                    if (pointer + 1 > [aData length]) break;
                    [aData getBytes:&b range:NSMakeRange(pointer,1)];
                    pointer++;

                    char b_v = b & 0x7F;
                    length = length * 128 + b_v;
                } while ((b & 0x80) == 0x80);

                if (length == 0) {
                    // Closing handshake frame
                    [self.delegate parserNeedsInitiateClosingHandshakeWithData:self.closingPayload];
                    finished = YES;
                } else if (length > [aData length] - pointer) {
                    // If there is not enough data in the frame it's incomplete frame
                    NSString *reason = [NSString stringWithFormat:@"We're supposed to read %d bytes but we only have %d bytes.",length,[aData length]];
                    NSDictionary *ui = [NSDictionary dictionaryWithObjectsAndKeys:reason, NSLocalizedDescriptionKey, nil];
                    NSError *error = [NSError errorWithDomain:PKWebSocketErrorDomain
                                                         code:PKWebSocketFrameIncompleteFrame
                                                     userInfo:ui];
                    [self.delegate parserDidProcessString:nil error:error];
                    finished = YES;
                } else {
                    // Set pointer to discard all dat to length
                    pointer += length;
                }
            } else {
                if (frameType != 0x00) {
                    NSString *reason = @"Text frame must start with 0x00.";
                    NSDictionary *ui = [NSDictionary dictionaryWithObjectsAndKeys:reason, NSLocalizedDescriptionKey, nil];
                    NSError *error = [NSError errorWithDomain:PKWebSocketErrorDomain
                                                         code:PKWebSocketFrameInvalidFrame
                                                     userInfo:ui];
                    [self.delegate parserDidProcessString:nil error:error];
                    finished = YES;
                } else {
                    // Text frame
                    // We read data till we reach 0xFF and copy that to the buffer
                    // keeping pointer in sync.
                    char byte;
                    NSMutableData *buffer = [NSMutableData dataWithCapacity:[aData length] - pointer];
                    NSRange range = NSMakeRange(pointer, 1);
                    [aData getBytes:&byte range:range];
                    pointer++;

                    while ((byte  & 0xFF) != 0xFF && pointer < [aData length]) {
                        [buffer appendBytes:&byte length:1];

                        NSRange range = NSMakeRange(pointer, 1);
                        [aData getBytes:&byte range:range];
                        pointer++;
                    }

                    // Convert data to UTF-8 String
                    NSError *error = nil;
                    NSString *text = [[NSString alloc] initWithData:buffer encoding:NSUTF8StringEncoding];
                    if (text == nil) {
                        NSString *reason = @"Unable to create UTF-8 string from data, missing 0xFF probably.";
                        NSDictionary *ui = [NSDictionary dictionaryWithObjectsAndKeys:reason, NSLocalizedDescriptionKey, nil];
                        error = [NSError errorWithDomain:PKWebSocketErrorDomain
                                                    code:PKWebSocketFrameInvalidFrame
                                                userInfo:ui];
                        finished = YES;
                    }
                    [self.delegate parserDidProcessString:text error:error];

                    // We reached end of data so we're finished
                    if (pointer == [aData length]) finished = YES;
                }
            }
        }
    }
}

@end

