//
//  PKWebSocket.m
//  PKWebSocket
//
//  This code is heavily based on Zimt (https://github.com/esad/zimt) by Esad 
//  Hajdarevic on 2/14/10.
//
//  Created by Pavel Kunc on 16/04/2012.
//  Copyright (c) 2012 Pavel Kunc. All rights reserved.
//

#import "PKWebSocket.h"
#import "PKWebSocketHandler.h"
#import "PKHTTPRequest.h"
#import "PKHTTPResponse.h"
#import "PKWebSocketMessage.h"
#import "PKWebSocketURL.h"


@implementation PKWebSocket

@synthesize delegate = _delegate;
@synthesize handler  = _handler;


#pragma mark - Initializers

- (id)initWithHandler:(PKWebSocketHandler *)aHandler
            delegate:(id<PKWebSocketDelegate>)aDelegate {
    NSParameterAssert(aHandler);

    self = [super init];
    if (self != nil) {
        self.handler  = aHandler;
        self.delegate = aDelegate;
        self.handler.delegate = self;
    }
    return self;
}


#pragma mark Public interface

- (void)close {
    [self.handler disconnect];
}

- (BOOL)openWithURLString:(NSString *)aURL error:(NSError **)outError {
    NSParameterAssert(aURL);

    PKWebSocketURL *url = [PKWebSocketURL URLWithString:aURL];
    if (url == nil) {
        if (outError != NULL) {
            NSString *reason = [NSString stringWithFormat:@"Passed URL (%@) is not valid WebSocket URL.", aURL];
            NSDictionary *ui = [NSDictionary dictionaryWithObjectsAndKeys:reason, NSLocalizedDescriptionKey, nil];
            *outError = [NSError errorWithDomain:PKWebSocketErrorDomain
                                            code:PKWebSocketInvalidURL
                                        userInfo:ui];
        }
        return NO;
    }
    return [self.handler connectWithURL:url error:outError];
}

- (void)send:(PKWebSocketMessage *)aMessage {
    NSParameterAssert(aMessage);

    if ([aMessage isBinary]) {
        [self sendData:aMessage.data];
    } else {
        [self sendString:aMessage.string];
    }
}

- (void)sendString:(NSString *)aString {
    NSParameterAssert(aString);
    [self.handler sendTextFrameWithString:aString];
}

- (void)sendData:(NSData *)aData {
    NSParameterAssert(aData);
    [self.handler sendBinaryFrameWithData:aData];
}


#pragma mark - Accessors

- (void)setTLSSettings:(NSDictionary *)aSettings {
    self.handler.tlsSettings = aSettings;
}

- (NSDictionary *)TLSSettings {
    return self.handler.tlsSettings;
}

- (PKWebSocketURL *)url {
    return self.handler.url;
}

- (void)setTimeout:(NSUInteger)aTimeout {
    self.handler.timeout = aTimeout;
}

- (NSUInteger)timeout {
    return self.handler.timeout;
}


#pragma mark Delegate dispatch methods

- (void)handler:(PKWebSocketHandler *)handler didChangeStatus:(PKWebSocketStatus)status {
    if (_delegate == nil) return;

    SEL sel = NULL;
    switch (status) {
        case PKWebSocketClosedStatus:
            sel = @selector(webSocketDidClose:);
            break;
        case PKWebSocketOpenedStatus:
            sel = @selector(webSocketDidOpen:);
            break;
        default:
            break;
    }
    if (sel != NULL && [_delegate respondsToSelector:sel]) {
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Warc-performSelector-leaks"
        [_delegate performSelector:sel withObject:self];
#pragma clang diagnostic pop
    }
}

- (void)handler:(PKWebSocketHandler *)handler didFailWithError:(NSError *)error {
    if(_delegate && [_delegate respondsToSelector:@selector(webSocket:didFailWithError:)]) {
        [_delegate webSocket:self didFailWithError:error];
    }
}

- (void)handler:(PKWebSocketHandler *)handler didProcessMessage:(PKWebSocketMessage *)message error:(NSError *)error {
    if (_delegate && [_delegate respondsToSelector:@selector(webSocket:didReceiveMessage:)]) {
        [_delegate webSocket:self didReceiveMessage:message];
    }
}

- (void)handlerDidSendMessage {
    if (_delegate && [_delegate respondsToSelector:@selector(webSocketDidSendMessage:)]) {
        [_delegate webSocketDidSendMessage:self];
    }
}

@end
