//
//  PKWebSocketHandlerTests.m
//  PKWebSocket
//
//  Created by Pavel Kunc on 17/04/2012.
//  Copyright (c) 2012 Pavel Kunc. All rights reserved.
//

#import <Foundation/Foundation.h>
#import <SenTestingKit/SenTestingKit.h>
#import "PKWebSocketHandler.h"
#import "PKWebSocketURL.h"
#import "PKHTTPRequest.h"
#import "PKHTTPResponse.h"


@interface PKWebSocketHandlerTests : SenTestCase
@end

@implementation PKWebSocketHandlerTests {
    PKWebSocketHandler *_handler;
}

- (void)setUp {
    [super setUp];

    _handler = [[PKWebSocketHandler alloc] init];
}

- (void)tearDown {
    _handler = nil;

    [super tearDown];
}


#pragma mark - Helpers

- (NSString *)validHandshakeResponse {
    return @"HTTP/1.1 101 Web Socket Protocol Handshake\r\n"
            "Upgrade: WebSocket\r\n"
            "Connection: Upgrade\r\n"
            "\r\n";
}

- (NSString *)wrongStatusCodeHandshakeResponse {
    return @"HTTP/1.1 404 Not Found\r\n"
            "\r\n";
}

- (NSString *)missingConnectionHeaderHandshakeResponse {
    return @"HTTP/1.1 101 Web Socket Protocol Handshake\r\n"
            "Upgrade: WebSocket\r\n"
            "\r\n";
}

- (NSString *)missingUpgradeHeaderHandshakeResponse {
    return @"HTTP/1.1 101 Web Socket Protocol Handshake\r\n"
            "Connection: Upgrade\r\n"
            "\r\n";
}

- (NSString *)noHeadersHandshakeResponse {
    return @"HTTP/1.1 101 Web Socket Protocol Handshake\r\n"
            "\r\n";
}

- (PKWebSocketURL *)URL {
    return [PKWebSocketURL URLWithString:@"ws://example.com/demo"];
}

- (PKWebSocketURL *)secureURL {
    return [PKWebSocketURL URLWithString:@"wss://example.com/demo"];
}


#pragma mark - Tests opening handshake

- (void)testCreateVaidOpeningHandshakeRequest {
    PKHTTPRequest *request = [_handler handshakeRequestWithURL:[self URL] error:NULL];
    STAssertNotNil(request, nil);
    STAssertEqualObjects([request valueForHTTPHeaderField:@"Upgrade"], @"WebSocket", nil);
    STAssertEqualObjects([request valueForHTTPHeaderField:@"Connection"], @"Upgrade", nil);
    STAssertNotNil([request valueForHTTPHeaderField:@"Origin"],nil);
    STAssertNotNil([request valueForHTTPHeaderField:@"Host"],nil);
}

- (void)testAcceptsCorrectOpeningHandshakeResponse {
    NSData *data = [[self validHandshakeResponse] dataUsingEncoding:NSASCIIStringEncoding];
    PKHTTPResponse *response = [[PKHTTPResponse alloc] initWithData:data error:NULL];
    STAssertTrue([_handler validateHandshakeResponse:response error:NULL], nil);
}

- (void)testDeclinesResponseWithInvalidStatusCode {
    NSError *error;
    NSData *data = [[self wrongStatusCodeHandshakeResponse] dataUsingEncoding:NSASCIIStringEncoding];
    PKHTTPResponse *response = [[PKHTTPResponse alloc] initWithData:data error:&error];
    STAssertFalse([_handler validateHandshakeResponse:response error:&error], nil);
}

- (void)testDeclinesResponseWithoutUpgradeHeader {
    NSError *error;
    NSData *data = [[self missingUpgradeHeaderHandshakeResponse] dataUsingEncoding:NSASCIIStringEncoding];
    PKHTTPResponse *response = [[PKHTTPResponse alloc] initWithData:data error:&error];
    STAssertFalse([_handler validateHandshakeResponse:response error:&error], nil);
}

- (void)testDeclinesResponseWithoutConnectionHeader {
    NSError *error;
    NSData *data = [[self missingConnectionHeaderHandshakeResponse] dataUsingEncoding:NSASCIIStringEncoding];
    PKHTTPResponse *response = [[PKHTTPResponse alloc] initWithData:data error:&error];
    STAssertFalse([_handler validateHandshakeResponse:response error:&error], nil);
}

- (void)testDeclinesResponseWithoutAnyHeaders {
    NSError *error;
    NSData *data = [[self noHeadersHandshakeResponse] dataUsingEncoding:NSASCIIStringEncoding];
    PKHTTPResponse *response = [[PKHTTPResponse alloc] initWithData:data error:&error];
    STAssertFalse([_handler validateHandshakeResponse:response error:&error], nil);
}

- (void)testDefaultHandshakeResponseBoundaryShouldBeTwoNewLines {
    NSData *expected = [@"\r\n\r\n" dataUsingEncoding:NSASCIIStringEncoding];
    STAssertTrue([[_handler handshakeResponseFrameBoundary] isEqualToData:expected],nil);
}

@end

